package gg.pufferfish.pufferfish.util;

import com.google.common.collect.Queues;
import gg.pufferfish.pufferfish.PufferfishLogger;
import java.util.Queue;
import java.util.concurrent.locks.LockSupport;
import java.util.function.BooleanSupplier;
import java.util.logging.Level;

public class AsyncExecutor implements Runnable {
	
	private Queue<Runnable> jobs = Queues.newConcurrentLinkedQueue();
	private final Thread thread;
	private final BooleanSupplier shouldRun;
	private volatile boolean killswitch = false;
	
	public AsyncExecutor(String threadName, BooleanSupplier shouldRun) {
		this.thread = new Thread(this, threadName);
		this.shouldRun = shouldRun;
	}
	
	public void start() {
		thread.start();
	}
	
	public void kill() {
		killswitch = true;
	}
	
	public void submit(Runnable runnable) {
		jobs.offer(runnable);
	}
	
	@Override
	public void run() {
		while (!killswitch) {
			if (shouldRun.getAsBoolean()) {
				try {
					Runnable runnable;
					while ((runnable = jobs.poll()) != null) {
						runnable.run();
					}
				} catch (Exception e) {
					PufferfishLogger.LOGGER.log(Level.SEVERE, e, () -> "Failed to execute async job for thread " + thread.getName());
				}
			}
			LockSupport.parkNanos("executing tasks", 1000L);
		}
	}
	
}
